import logging
from typing import List, Optional

from langfoundation.modelhub.rerank.config import RerankConfig


logger = logging.getLogger(__name__)

try:
    from llama_index.core.callbacks import CBEventType, EventPayload  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.postprocessor.types import BaseNodePostprocessor  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.schema import NodeWithScore, QueryBundle  # type: ignore[unused-ignore] # noqa: F401
except ImportError:
    raise ImportError()


class LlamaIndexCrossEncoder(BaseNodePostprocessor):
    config: RerankConfig
    keep_retrieval_score: bool = True
    keep_rerank_score: bool = True

    def _postprocess_nodes(
        self,
        nodes: List[NodeWithScore],
        query_bundle: Optional[QueryBundle] = None,
    ) -> List[NodeWithScore]:
        if query_bundle is None:
            raise ValueError("Missing query bundle in extra info.")
        if len(nodes) == 0:
            return []

        query = query_bundle.query_str
        # TODO: To test LLM, embedding 🚜
        docs = [node.get_text() for node in nodes]
        model = self.config.model()
        top_n = self.config.top_n

        with self.callback_manager.event(
            CBEventType.RERANKING,
            payload={
                EventPayload.NODES: nodes,
                EventPayload.MODEL_NAME: self.config.model_provider,
                EventPayload.QUERY_STR: query,
                EventPayload.TOP_K: self.config.top_n,
            },
        ) as event:
            indexes_scores = model.rerank(query=query, docs=docs)
            # Incase response not sorted and not remove top_n
            indexes_scores = sorted(indexes_scores, key=lambda x: x[1], reverse=True)[:top_n]

            sorted_nodes = []
            for index, score in indexes_scores:
                node = nodes[index]

                if self.keep_retrieval_score:
                    node.node.metadata["retrieval_score"] = nodes[index].score
                if self.keep_rerank_score:
                    node.node.metadata["rerank_score"] = score

                sorted_nodes.append(node)
            event.on_end(payload={EventPayload.NODES: sorted_nodes})

        return sorted_nodes

    @classmethod
    def class_name(cls) -> str:
        return "SentenceTransformerRerank"



from functools import lru_cache
import logging
from typing import Any, List, Tuple, Type, Union

from langchain_community.cross_encoders import BaseCrossEncoder

from langfoundation.modelhub.rerank.base import BaseRerankModel, BaseRerankProvider


try:
    from langchain_community.cross_encoders import HuggingFaceCrossEncoder
except ImportError:
    raise ImportError()


logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _get_cross_encoder_instance(
    model_name: str,
    provider: Union[
        Type[BaseCrossEncoder],
        Type[Any],
    ],
) -> BaseCrossEncoder:
    config = {
        "model_name": model_name,
        "model_kwargs": {
            "trust_remote_code": True,
        },
    }
    logger.info(
        config,
        extra={
            "title": "[Cache Model] " + " : " + model_name,
            "verbose": True,
        },
    )
    return provider(**config)


class RerankCrossEncoder(BaseRerankModel):
    """Interface for cross encoder models."""

    model_name: str
    provider: Union[
        Type[BaseCrossEncoder],
        Type[Any],
    ]

    @property
    def _model(
        self,
    ) -> BaseCrossEncoder:
        return _get_cross_encoder_instance(
            model_name=self.model_name,
            provider=self.provider,
        )

    def rerank(
        self,
        query: str,
        docs: List[str],
    ) -> List[Tuple[int, float]]:
        query_and_nodes = [
            (
                query,
                doc,
            )
            for doc in docs
        ]
        scores = self._model.score(query_and_nodes)

        indexs_scores = [(index, score) for index, score in enumerate(scores)]

        return sorted(indexs_scores, key=lambda x: -x[1] if x[1] else 0)[: self.top_n]


class HuggingFaceRerankProvider(BaseRerankProvider):
    BGE_RERANKER_LAGER = "BAAI/bge-reranker-large"

    @property
    def provider(
        self,
    ) -> Union[
        Type[BaseCrossEncoder],
        Type[Any],
    ]:
        return HuggingFaceCrossEncoder

    def model(
        self,
        top_n: int,
    ) -> BaseRerankModel:
        return RerankCrossEncoder(
            top_n=top_n,
            model_name=self.value,
            provider=self.provider,
        )



from typing import Dict, List, Sequence


try:
    from llama_index.core.base.llms.types import ChatMessage, LLMMetadata, MessageRole  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.llms.openai.utils import openai_modelname_to_contextsize  # type: ignore[unused-ignore] # noqa: F401
except ImportError:
    raise ImportError("Please install `llama_index` library. Please install `poetry add 'langcommunity[llama_index]'`.")


DEFAULT_ANYSCALE_API_BASE = "https://api.endpoints.anyscale.com/v1"
DEFAULT_ANYSCALE_API_VERSION = ""

LLAMA_MODELS = {
    "meta-llama/Llama-2-7b-chat-hf": 4096,
    "meta-llama/Llama-2-13b-chat-hf": 4096,
    "meta-llama/Llama-2-70b-chat-hf": 4096,
    "codellama/CodeLlama-34b-Instruct-hf": 16384,
    "Meta-Llama/Llama-Guard-7b": 4096,
}

MISTRAL_MODELS = {
    "mistralai/Mistral-7B-Instruct-v0.1": 16384,
    "Open-Orca/Mistral-7B-OpenOrca": 8192,
    "mistralai/Mixtral-8x7B-Instruct-v0.1": 32768,
}

ZEPHYR_MODELS = {
    "HuggingFaceH4/zephyr-7b-beta": 16384,
}

ALL_AVAILABLE_MODELS = {
    **LLAMA_MODELS,
    **MISTRAL_MODELS,
    **ZEPHYR_MODELS,
}

DISCONTINUED_MODELS: Dict[str, int] = {}


def anyscale_modelname_to_contextsize(modelname: str) -> int:
    """
    Calculate the maximum number of tokens possible to generate for a model.

    Args:
        modelname: The modelname we want to know the context size for.

    Returns:
        The maximum context size

    Example:
        .. code-block:: python

            max_tokens = anyscale_modelname_to_contextsize(model_name)
    """
    # handling finetuned models
    # TO BE FILLED

    if modelname in DISCONTINUED_MODELS:
        raise ValueError(f"Anyscale hosted model {modelname} has been discontinued. " "Please choose another model.")

    context_size = ALL_AVAILABLE_MODELS.get(modelname, None)

    if context_size is None:
        raise ValueError(
            f"Unknown model: {modelname}. Please provide a valid Anyscale model name."
            "Known models are: " + ", ".join(ALL_AVAILABLE_MODELS.keys())
        )

    return context_size


class LC:
    from llama_index.core.bridge.langchain import (  # AI21,; Cohere,; OpenAI,
        AIMessage,
        BaseChatModel,
        BaseLanguageModel,
        BaseMessage,
        ChatAnyscale,
        ChatFireworks,
        ChatMessage,
        ChatOpenAI,
        FunctionMessage,
        HumanMessage,
        SystemMessage,
    )


def is_chat_model(llm: LC.BaseLanguageModel) -> bool:
    return isinstance(llm, LC.BaseChatModel)


def to_lc_messages(messages: Sequence[ChatMessage]) -> List[LC.BaseMessage]:
    lc_messages: List[LC.BaseMessage] = []
    for message in messages:
        LC_MessageClass = LC.BaseMessage
        lc_kw = {
            "content": message.content,
            "additional_kwargs": message.additional_kwargs,
        }
        if message.role == "user":
            LC_MessageClass = LC.HumanMessage
        elif message.role == "assistant":
            LC_MessageClass = LC.AIMessage
        elif message.role == "function":
            LC_MessageClass = LC.FunctionMessage
        elif message.role == "system":
            LC_MessageClass = LC.SystemMessage
        elif message.role == "chatbot":
            LC_MessageClass = LC.ChatMessage
        else:
            raise ValueError(f"Invalid role: {message.role}")

        for req_key in LC_MessageClass.schema().get("required"):
            if req_key not in lc_kw:
                more_kw = lc_kw.get("additional_kwargs")
                if not isinstance(more_kw, dict):
                    raise ValueError(f"additional_kwargs must be a dict, got {type(more_kw)}")
                if req_key not in more_kw:
                    raise ValueError(f"{req_key} needed for {LC_MessageClass}")
                lc_kw[req_key] = more_kw.pop(req_key)

        lc_messages.append(LC_MessageClass(**lc_kw))

    return lc_messages


def from_lc_messages(lc_messages: Sequence[LC.BaseMessage]) -> List[ChatMessage]:
    messages: List[ChatMessage] = []
    for lc_message in lc_messages:
        li_kw = {
            "content": lc_message.content,
            "additional_kwargs": lc_message.additional_kwargs,
        }
        if isinstance(lc_message, LC.HumanMessage):
            li_kw["role"] = MessageRole.USER
        elif isinstance(lc_message, LC.AIMessage):
            li_kw["role"] = MessageRole.ASSISTANT
        elif isinstance(lc_message, LC.FunctionMessage):
            li_kw["role"] = MessageRole.FUNCTION
        elif isinstance(lc_message, LC.SystemMessage):
            li_kw["role"] = MessageRole.SYSTEM
        elif isinstance(lc_message, LC.ChatMessage):
            li_kw["role"] = MessageRole.CHATBOT
        else:
            raise ValueError(f"Invalid message type: {type(lc_message)}")
        messages.append(ChatMessage(**li_kw))

    return messages


def get_llm_metadata(llm: LC.BaseLanguageModel) -> LLMMetadata:
    """Get LLM metadata from llm."""
    if not isinstance(llm, LC.BaseLanguageModel):
        raise ValueError("llm must be instance of LangChain BaseLanguageModel")

    is_chat_model_ = is_chat_model(llm)

    # if isinstance(llm, LC.OpenAI):
    #     return LLMMetadata(
    #         context_window=openai_modelname_to_contextsize(llm.model_name),
    #         num_output=llm.max_tokens,
    #         is_chat_model=is_chat_model_,
    #         model_name=llm.model_name,
    #     )
    if isinstance(llm, LC.ChatAnyscale):
        return LLMMetadata(
            context_window=anyscale_modelname_to_contextsize(llm.model_name),
            num_output=llm.max_tokens or -1,
            is_chat_model=is_chat_model_,
            model_name=llm.model_name,
        )
    # elif isinstance(llm, LC.ChatFireworks):
    #     return LLMMetadata(
    #         context_window=fireworks_modelname_to_contextsize(llm.model_name),
    #         num_output=llm.max_tokens or -1,
    #         is_chat_model=is_chat_model_,
    #         model_name=llm.model_name,
    #     )
    elif isinstance(llm, LC.ChatOpenAI):
        return LLMMetadata(
            context_window=openai_modelname_to_contextsize(llm.model_name),
            num_output=llm.max_tokens or -1,
            is_chat_model=is_chat_model_,
            model_name=llm.model_name,
        )
    # elif isinstance(llm, LC.Cohere):
    #     # June 2023: Cohere's supported max input size for Generation models is 2048
    #     # Reference: <https://docs.cohere.com/docs/tokens>
    #     return LLMMetadata(
    #         context_window=COHERE_CONTEXT_WINDOW,
    #         num_output=llm.max_tokens,
    #         is_chat_model=is_chat_model_,
    #         model_name=llm.model,
    #     )
    # elif isinstance(llm, LC.AI21):
    #     # June 2023:
    #     #   AI21's supported max input size for
    #     #   J2 models is 8K (8192 tokens to be exact)
    #     # Reference: <https://docs.ai21.com/changelog/increased-context-length-for-j2-foundation-models>
    #     return LLMMetadata(
    #         context_window=AI21_J2_CONTEXT_WINDOW,
    #         num_output=llm.maxTokens,
    #         is_chat_model=is_chat_model_,
    #         model_name=llm.model,
    #     )
    else:
        return LLMMetadata(is_chat_model=is_chat_model_)



from typing import Any, Callable, Generator, Optional, Sequence

from langchain_core.language_models import BaseLanguageModel


try:
    from llama_index.core.base.llms.generic_utils import (  # type: ignore[unused-ignore] # noqa: F401
        completion_response_to_chat_response,
        stream_completion_response_to_chat_response,
    )
    from llama_index.core.base.llms.types import (  # type: ignore[unused-ignore] # noqa: F401
        ChatMessage,
        ChatResponse,
        ChatResponseAsyncGen,
        ChatResponseGen,
        CompletionResponse,
        CompletionResponseAsyncGen,
        CompletionResponseGen,
        LLMMetadata,
    )
    from llama_index.core.bridge.pydantic import PrivateAttr  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.callbacks import CallbackManager  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.llms.callbacks import llm_chat_callback, llm_completion_callback  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.llms.llm import LLM  # type: ignore[unused-ignore] # noqa: F401
    from llama_index.core.types import BaseOutputParser, PydanticProgramMode, Thread  # type: ignore[unused-ignore] # noqa: F401
    from modelhub.adapter.chat.utils import from_lc_messages, get_llm_metadata, to_lc_messages
except ImportError:
    raise ImportError("Please install `llama_index` library. Please install `poetry add llama-index`.")


class LlamaIndexLangChainChatModel(LLM):
    """Adapter for a LangChain LLM.

    Examples:
        `poetry add llama-index`

        ```python
        from langchain_openai import ChatOpenAI

        from llama_index.llms.langchain import LlamaIndexLangChainLLM

        llm = LlamaIndexLangChainLLM(llm=ChatOpenAI(...))

        response_gen = llm.stream_complete("What is the meaning of life?")

        for r in response_gen:
            print(r.delta, end="")
        ```
    """

    _llm: Any = PrivateAttr()

    def __init__(
        self,
        llm: "BaseLanguageModel",
        callback_manager: Optional[CallbackManager] = None,
        system_prompt: Optional[str] = None,
        messages_to_prompt: Optional[Callable[[Sequence[ChatMessage]], str]] = None,
        completion_to_prompt: Optional[Callable[[str], str]] = None,
        pydantic_program_mode: PydanticProgramMode = PydanticProgramMode.DEFAULT,
        output_parser: Optional[BaseOutputParser] = None,
    ) -> None:
        self._llm = llm
        super().__init__(
            callback_manager=callback_manager,
            system_prompt=system_prompt,
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            pydantic_program_mode=pydantic_program_mode,
            output_parser=output_parser,
        )

    @classmethod
    def class_name(cls) -> str:
        return "LlamaIndexLangChainLLM"

    @property
    def llm(self) -> "BaseLanguageModel":
        return self._llm

    @property
    def metadata(self) -> LLMMetadata:
        return get_llm_metadata(self._llm)

    @llm_chat_callback()  # type: ignore[unused-ignore] # noqa: F401
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        if not self.metadata.is_chat_model:
            prompt = self.messages_to_prompt(messages)
            completion_response = self.complete(prompt, formatted=True, **kwargs)
            return completion_response_to_chat_response(completion_response)

        lc_messages = to_lc_messages(messages)
        lc_message = self._llm.predict_messages(messages=lc_messages, **kwargs)
        message = from_lc_messages([lc_message])[0]
        return ChatResponse(message=message)

    @llm_completion_callback()  # type: ignore[unused-ignore] # noqa: F401
    def complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        if not formatted:
            prompt = self.completion_to_prompt(prompt)

        output_str = self._llm.predict(prompt, **kwargs)
        return CompletionResponse(text=output_str)

    @llm_chat_callback()  # type: ignore[unused-ignore] # noqa: F401
    def stream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseGen:
        if not self.metadata.is_chat_model:
            prompt = self.messages_to_prompt(messages)
            stream_completion = self.stream_complete(prompt, formatted=True, **kwargs)
            return stream_completion_response_to_chat_response(stream_completion)

        if hasattr(self._llm, "stream"):

            def gen() -> Generator[ChatResponse, None, None]:
                lc_messages = to_lc_messages(messages)
                response_str = ""
                for message in self._llm.stream(lc_messages, **kwargs):
                    message = from_lc_messages([message])[0]
                    delta = message.content
                    response_str += delta
                    yield ChatResponse(
                        message=ChatMessage(role=message.role, content=response_str),
                        delta=delta,
                    )

            return gen()

        else:
            from llama_index.core.langchain_helpers.streaming import (
                StreamingGeneratorCallbackHandler,  # type: ignore[unused-ignore] # noqa: F401
            )

            handler = StreamingGeneratorCallbackHandler()

            if not hasattr(self._llm, "streaming"):
                raise ValueError("LLM must support streaming.")
            if not hasattr(self._llm, "callbacks"):
                raise ValueError("LLM must support callbacks to use streaming.")

            self._llm.callbacks = [handler]  # type: ignore
            self._llm.streaming = True  # type: ignore

            thread = Thread(target=self.chat, args=[messages], kwargs=kwargs)
            thread.start()

            response_gen = handler.get_response_gen()

            def gen() -> Generator[ChatResponse, None, None]:
                text = ""
                for delta in response_gen:
                    text += delta
                    yield ChatResponse(
                        message=ChatMessage(text=text),
                        delta=delta,
                    )

            return gen()

    @llm_completion_callback()  # type: ignore[unused-ignore] # noqa: F401
    def stream_complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponseGen:
        if not formatted:
            prompt = self.completion_to_prompt(prompt)

        from llama_index.core.langchain_helpers.streaming import (
            StreamingGeneratorCallbackHandler,  # type: ignore[unused-ignore] # noqa: F401
        )

        handler = StreamingGeneratorCallbackHandler()

        if not hasattr(self._llm, "streaming"):
            raise ValueError("LLM must support streaming.")
        if not hasattr(self._llm, "callbacks"):
            raise ValueError("LLM must support callbacks to use streaming.")

        self._llm.callbacks = [handler]  # type: ignore
        self._llm.streaming = True  # type: ignore

        thread = Thread(target=self.complete, args=[prompt], kwargs=kwargs)
        thread.start()

        response_gen = handler.get_response_gen()

        def gen() -> Generator[CompletionResponse, None, None]:
            text = ""
            for delta in response_gen:
                text += delta
                yield CompletionResponse(delta=delta, text=text)

        return gen()

    @llm_chat_callback()  # type: ignore[unused-ignore] # noqa: F401
    async def achat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        # TODO: Implement async chat
        return self.chat(messages, **kwargs)

    @llm_completion_callback()
    async def acomplete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponse:
        # TODO: Implement async complete
        return self.complete(prompt, formatted=formatted, **kwargs)

    @llm_chat_callback()
    async def astream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseAsyncGen:
        # TODO: Implement async stream_chat

        async def gen() -> ChatResponseAsyncGen:
            for message in self.stream_chat(messages, **kwargs):
                yield message

        return gen()

    @llm_completion_callback()
    async def astream_complete(self, prompt: str, formatted: bool = False, **kwargs: Any) -> CompletionResponseAsyncGen:
        # TODO: Implement async stream_complete

        async def gen() -> CompletionResponseAsyncGen:
            for response in self.stream_complete(prompt, formatted=formatted, **kwargs):
                yield response

        return gen()



from typing import Any, Dict, Type

from langchain_core.embeddings import Embeddings
from langfoundation.modelhub.embedding.base import BaseEmbeddingProvider


try:
    from langchain_huggingface import HuggingFaceEmbeddings  # type: ignore
except ImportError:
    raise ImportError()


class HuggingFaceEmbeddingProvider(BaseEmbeddingProvider):
    GTE_LARGE_EN_V_1_5 = "Alibaba-NLP/gte-large-en-v1.5"

    @property
    def provider(self) -> Type[Embeddings]:
        return HuggingFaceEmbeddings

    @property
    def config(self) -> Dict[str, Any]:
        return {
            "model_name": self.value,
            "model_kwargs": {"trust_remote_code": True},
        }



import logging
from typing import Any, List, Tuple, Type, Union

from langfoundation.modelhub.rerank.base import BaseRerankModel


logger = logging.getLogger(__name__)


try:
    from cohere.client import Client as CohereClient  # type: ignore[unused-ignore] # noqa: F401
except ImportError:
    raise ImportError()


class RerankCohere(BaseRerankModel):
    """Interface for cross encoder models."""

    client: Union[
        Type[CohereClient],
        Type[Any],
    ]
    model_name: str

    def rerank(
        self,
        query: str,
        docs: List[str],
    ) -> List[Tuple[int, float]]:
        with self.client() as client:
            response = client.rerank(
                model=self.model_name,
                query=query,
                documents=docs,
                top_n=self.top_n,
            )

        return [(result.index, result.relevance_score) for result in response.results]



from typing import Any, Dict, Type

from langchain_core.embeddings import Embeddings
from langfoundation.modelhub.embedding.base import BaseEmbeddingProvider


try:
    from modelhub.adapter.langchain_overrides.mistral_embeddings import MistralAIEmbeddings  # type: ignore
except ImportError:
    raise ImportError()


class MistralEmbeddingProvider(BaseEmbeddingProvider):
    EMBED = "mistral-embed"

    @property
    def provider(self) -> Type[Embeddings]:
        return MistralAIEmbeddings

    @property
    def config(self) -> Dict[str, Any]:
        return {"model": self.value}



import asyncio
import logging
from typing import Any, Dict, Iterable, List, Optional

import httpx
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import BaseModel, Extra, Field, SecretStr, root_validator
from langchain_core.utils import convert_to_secret_str, get_from_dict_or_env
from tokenizers import Tokenizer  # type: ignore

logger = logging.getLogger(__name__)

MAX_TOKENS = 16_000
"""A batching parameter for the Mistral API. This is NOT the maximum number of tokens
accepted by the embedding model for each document/chunk, but rather the maximum number
of tokens that can be sent in a single request to the Mistral API (across multiple
documents/chunks)"""


class DummyTokenizer:
    """Dummy tokenizer for when tokenizer cannot be accessed (e.g., via Huggingface)"""

    def encode_batch(self, texts: List[str]) -> List[List[str]]:
        return [list(text) for text in texts]


# NOTE: Override class, du to not needed error with tokenizer
class MistralAIEmbeddings(BaseModel, Embeddings):
    """MistralAI embedding models.

    To use, set the environment variable `MISTRAL_API_KEY` is set with your API key or
    pass it as a named parameter to the constructor.

    Example:
        .. code-block:: python

            from langchain_mistralai import MistralAIEmbeddings

            mistral = MistralAIEmbeddings(
                model="mistral-embed",
                api_key="my-api-key"
            )
    """

    client: httpx.Client = Field(default=None)  #: :meta private:
    async_client: httpx.AsyncClient = Field(default=None)  #: :meta private:
    mistral_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    endpoint: str = "https://api.mistral.ai/v1/"
    max_retries: int = 5
    timeout: int = 120
    max_concurrent_requests: int = 64
    tokenizer: Tokenizer = Field(default_factory=DummyTokenizer)

    model: str = "mistral-embed"

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        allow_population_by_field_name = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate configuration."""

        values["mistral_api_key"] = convert_to_secret_str(get_from_dict_or_env(values, "mistral_api_key", "MISTRAL_API_KEY", default=""))
        api_key_str = values["mistral_api_key"].get_secret_value()
        # todo: handle retries
        if not values.get("client"):
            values["client"] = httpx.Client(
                base_url=values["endpoint"],
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "Authorization": f"Bearer {api_key_str}",
                },
                timeout=values["timeout"],
            )
        # todo: handle retries and max_concurrency
        if not values.get("async_client"):
            values["async_client"] = httpx.AsyncClient(
                base_url=values["endpoint"],
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "Authorization": f"Bearer {api_key_str}",
                },
                timeout=values["timeout"],
            )
        # if values["tokenizer"] is None:
        #     try:
        #         values["tokenizer"] = Tokenizer.from_pretrained("mistralai/Mixtral-8x7B-v0.1")
        #     except IOError:  # huggingface_hub GatedRepoError
        #         warnings.warn(
        #             "Could not download mistral tokenizer from Huggingface for "
        #             "calculating batch sizes. Set a Huggingface token via the "
        #             "HF_TOKEN environment variable to download the real tokenizer. "
        #             "Falling back to a dummy tokenizer that uses `len()`."
        #         )
        #         values["tokenizer"] = DummyTokenizer()
        return values

    def _get_batches(self, texts: List[str]) -> Iterable[List[str]]:
        """Split a list of texts into batches of less than 16k tokens
        for Mistral API."""
        batch: List[str] = []
        batch_tokens = 0

        text_token_lengths = [len(encoded) for encoded in self.tokenizer.encode_batch(texts)]

        for text, text_tokens in zip(texts, text_token_lengths):
            if batch_tokens + text_tokens > MAX_TOKENS:
                if len(batch) > 0:
                    # edge case where first batch exceeds max tokens
                    # should not yield an empty batch.
                    yield batch
                batch = [text]
                batch_tokens = text_tokens
            else:
                batch.append(text)
                batch_tokens += text_tokens
        if batch:
            yield batch

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of document texts.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        try:
            batch_responses = (
                self.client.post(
                    url="/embeddings",
                    json=dict(
                        model=self.model,
                        input=batch,
                    ),
                )
                for batch in self._get_batches(texts)
            )
            return [
                list(map(float, embedding_obj["embedding"])) for response in batch_responses for embedding_obj in response.json()["data"]
            ]
        except Exception as e:
            logger.error(f"An error occurred with MistralAI: {e}")
            raise

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of document texts.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        max_retries = 3
        retry_interval = 0.5

        async def embed_batch(batch: List[str]) -> Any:
            for attempt in range(max_retries):
                response: httpx.Response
                try:
                    response = await self.async_client.post(
                        url="/embeddings",
                        json=dict(
                            model=self.model,
                            input=batch,
                        ),
                    )
                    return response.json()["data"]
                except Exception as e:
                    if attempt < max_retries - 1:
                        logger.error(f"An error occurred with MistralAI Retry, Got: {e}, Reponse: {response}")
                        await asyncio.sleep(retry_interval)
                    else:
                        logger.error(f"An error occurred with MistralAI Fail, Got: {e}, Reponse: {response}")
                        raise

        batches = self._get_batches(texts)
        tasks = [embed_batch(batch) for batch in batches]

        try:
            batch_responses = await asyncio.gather(*tasks)
            return [list(map(float, embedding_obj["embedding"])) for response in batch_responses for embedding_obj in response]
        except Exception as e:
            logger.error(f"An error occurred with MistralAI: {e}")
            raise

    def embed_query(self, text: str) -> List[float]:
        """Embed a single query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        return self.embed_documents([text])[0]

    async def aembed_query(self, text: str) -> List[float]:
        """Embed a single query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        return (await self.aembed_documents([text]))[0]
